# -*- coding: utf-8 -*-
# uncertainties_utilities.py provides additional functions uncertainty analysis
# Micha Wildermuth, micha.wildermuth@kit.edu 2023

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
import uncertainties as uncert
from uncertainties import ufloat, unumpy as unp


def sem(a, **kwargs):
    """
    Compute the standard error of the mean along the specified axis.

    Returns the standard error of the mean, that is per definition the standard deviation divided by sqrt(n-1). The
    standard error of the mean is computed for the
    flattened array by default, otherwise over the specified axis.

    Parameters
    ----------
    a : array_like
        Calculate the standard error of the mean of these values.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the standard error of the mean is computed. The default is to compute the standard
        error of the mean of the flattened array.

        If this is a tuple of ints, a standard error of the mean is performed over multiple axes, instead of a single
        axis or all the axes as before.
    dtype : dtype, optional
        Type to use in computing the standard error of the mean. For arrays of integer type the default is float64, for
        arrays of float types it is the same as the array type.
    out : ndarray, optional
        Alternative output array in which to place the result. It must have the same shape as the expected output but
        the type (of the calculated values) will be cast if necessary.
    ddof : int, optional
        Means Delta Degrees of Freedom.  The divisor used in calculations is ``N - ddof``, where ``N`` represents the
        number of elements. By default `ddof` is zero.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the result as dimensions with size one. With this
        option, the result will broadcast correctly against the input array.

        If the default value is passed, then `keepdims` will not be passed through to the `std` method of subclasses of
        `ndarray`, however any non-default value will be. If the subclass' method does not implement `keepdims` any
        exceptions will be raised.

    Returns
    -------
    sem : ndarray, see dtype parameter above.
        If `out` is None, return a new array containing the standard error of the mean, otherwise return a reference to
        the output array.

    References
    ----------
    https://en.wikipedia.org/wiki/Standard_error
    """
    n = a.shape[kwargs['axis']] if 'axis' in kwargs else np.prod(a.shape)
    return np.std(unp.nominal_values(a), **kwargs) / np.sqrt(n - 1)


def nansem(a, **kwargs):
    """
    Compute the standard error of the mean along the specified axis, while ignoring NaNs.

    Returns the standard error of the mean, that is per definition the standard deviation divided by sqrt(n-1), of the
    non-NaN array elements. The standard error of the mean is computed for the flattened array by default, otherwise
    over the specified axis.

    For all-NaN slices or slices with zero degrees of freedom, NaN is returned and a `RuntimeWarning` is raised.

    Parameters
    ----------
    a : array_like
        Calculate the standard error of the mean of the non-NaN values.
    axis : {int, tuple of int, None}, optional
        Axis or axes along which the standard error of the mean is computed. The default is to compute the standard
        error of the mean of the flattened array.
    dtype : dtype, optional
        Type to use in computing the standard error of the mean. For arrays of integer type the default is float64, for
        arrays of float types it is the same as the array type.
    out : ndarray, optional
        Alternative output array in which to place the result. It must have the same shape as the expected output but
        the type (of the calculated values) will be cast if necessary.
    ddof : int, optional
        Means Delta Degrees of Freedom.  The divisor used in calculations is ``N - ddof``, where ``N`` represents the
        number of non-NaN elements. By default `ddof` is zero.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the result as dimensions with size one. With this
        option, the result will broadcast correctly against the original `a`.

        If this value is anything but the default it is passed through as-is to the relevant functions of the
        subclasses. If these functions do not have a `keepdims` kwarg, a RuntimeError will be raised.

    Returns
    -------
    sem : ndarray, see dtype parameter above.
        If `out` is None, return a new array containing the standard error of the mean, otherwise return a reference to
        the output array. If ddof is >= the number of non-NaN elements in a slice or the slice contains only NaNs, then
        the result for that slice is NaN.

    References
    ----------
    https://en.wikipedia.org/wiki/Standard_error
    """
    n = a.shape[kwargs['axis']] if 'axis' in kwargs else np.prod(a.shape)
    return np.nanstd(unp.nominal_values(a), **kwargs) / np.sqrt(n - 1)


def umean(a, **kwargs):
    """
    Compute the arithmetic mean and its uncertainties of numbers with uncertainties along the specified axis.

    Returns the average and its uncertainty of the array elements. The average is taken over the flattened array by
    default, otherwise over the specified axis.
    The uncertainty is calculated as standard error of the mean as well as the error propagated uncertainties of the
    input array.
    `float64` intermediate and return values are used for integer inputs.

    Parameters
    ----------
    a : uncertainties.unumpy.ndarray
        Array containing numbers with uncertainties whose mean is desired. If `a` is not an array, a conversion is
        attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the means are computed. The default is to compute the mean of the flattened array.

        If this is a tuple of ints, a mean is performed over multiple axes, instead of a single axis or all the axes as
        before.
    dtype : data-type, optional
        Type to use in computing the mean. For integer inputs, the default is `float64`; for floating point inputs, it
        is the same as the input dtype.
    out : ndarray, optional
        Alternate output array in which to place the result. The default is ``None``; if provided, it must have the same
        shape as the expected output, but the type will be cast if necessary.
        See `ufuncs-output-type` for more details.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the result as dimensions with size one. With this
        option, the result will broadcast correctly against the input array.

        If the default value is passed, then `keepdims` will not be passed through to the `mean` method of subclasses
        of `ndarray`, however any non-default value will be. If the subclass' method does not implement `keepdims` any
        exceptions will be raised.

    Returns
    -------
    m : uncertainties.unumpy.ndarray
        If `out=None`, returns a new array containing the mean values and its uncertainties, otherwise a reference to
        the output array is returned.
    """
    avg = np.float64(np.squeeze(unp.nominal_values(np.mean(a, **kwargs))))
    std = np.float64(np.squeeze(np.sqrt(unp.std_devs(np.mean(a, **kwargs)) ** 2 +
                                        sem(unp.nominal_values(a), **kwargs) ** 2)))
    if isinstance(avg, np.floating) and isinstance(std, np.floating):
        return ufloat(nominal_value=avg,
                      std_dev=std)
    elif isinstance(avg, np.ndarray) and isinstance(std, np.ndarray):
        return unp.uarray(nominal_values=avg,
                          std_devs=std)
    else:
        raise TypeError('average value and standard deviation do not have compatible dtypes.')


def unanmean(a, **kwargs):
    """
    Compute the arithmetic mean and its uncertainties of numbers with uncertainties along the specified axis, ignoring
    NaNs.

    Returns the average and its uncertainty of the array elements. The average is taken over the flattened array by
    default, otherwise over the specified axis. The uncertainty is calculated as standard error of the mean as well as
    the error propagated uncertainties of the input array.
    `float64` intermediate and return values are used for integer inputs.

    For all-NaN slices, NaN is returned and a `RuntimeWarning` is raised.

    Parameters
    ----------
    a : uncertainties.unumpy.ndarray
        Array containing numbers with uncertainties whose mean is desired. If `a` is not an array, a conversion is
        attempted.
    axis : {int, tuple of int, None}, optional
        Axis or axes along which the means are computed. The default is to compute the mean of the flattened array.
        dtype : data-type, optional
        Type to use in computing the mean. For integer inputs, the default is `float64`; for inexact inputs, it is the
        same as the input dtype.
    out : ndarray, optional
        Alternate output array in which to place the result. The default is ``None``; if provided, it must have the same
        shape as the expected output, but the type will be cast if necessary.
        See `ufuncs-output-type` for more details.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the result as dimensions with size one. With this
        option, the result will broadcast correctly against the original `a`.
        If the value is anything but the default, then `keepdims` will be passed through to the `mean` or `sum` methods
        of subclasses of `ndarray`. If the subclasses methods does not implement `keepdims` any exceptions will be raised.

    Returns
    -------
    retval : uncertainties.unumpy.ndarray, see dtype parameter above
        If `out=None`, returns a new array containing the mean values and its uncertainties, otherwise a reference to
        the output array is returned. Nan is returned for slices that contain only NaNs.
    """
    avg = np.float64(np.squeeze(unp.nominal_values(np.nanmean(a, **kwargs))))
    std = np.float64(np.squeeze(np.sqrt(unp.std_devs(np.nanmean(a, **kwargs)) ** 2 +
                                        nansem(unp.nominal_values(a), **kwargs) ** 2)))
    if isinstance(avg, np.floating) and isinstance(std, np.floating):
        return ufloat(nominal_value=avg,
                      std_dev=std)
    elif isinstance(avg, np.ndarray) and isinstance(std, np.ndarray):
        return unp.uarray(nominal_values=avg,
                          std_devs=std)
    else:
        raise TypeError('average value and standard deviation do not have compatible dtypes.')


def uaverage(a, axis=None):
    """
    Computes the average and its uncertainties of numbers with uncertainties along the specified axis considering the
    uncertainties of each input value.

    Returns the average and its uncertainty of the array elements weighted by the inverse variance of each input value
    [1]_ that results from the maximum likelihood estimator applied to aGaussian distribution. The average is calculated
    as weighted arithmetic mean [2]_ and the uncertainty is calculated as square root of the weighted sample variance of
    the input array [3]_.
    Both weighted arithmetic mean and weighted sample variance is taken over the flattened array by default, otherwise
    over the specified axis.

    Parameters
    ----------
    a : uncertainties.unumpy.ndarray
        Array containing data with uncertainties to be averaged. If `a` is not an array, a conversion is attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which to average `a`.  The default, axis=None, will average over all the elements of the
        input array.
        If axis is negative it counts from the last to the first axis.
        If axis is a tuple of ints, averaging is performed on all the axes specified in the tuple instead of a single
        axis or all the axes as before.

    Returns
    -------
    retval : uncertainties.unumpy.ndarray or uncertainties.ufloat
        Returns the average and its uncertainty along the specified axis weighted by the inverse variance of each input
        value.
        The result dtype follows a general pattern and is uncertainties.ufloat if `axis` is None or else
        uncertainties.unumpy.ndarray.

    References
    ----------
    [1] Guide to the expression of uncertainty in measurement: https://www.bipm.org/utils/common/documents/jcgm/JCGM_100_2008_E.pdf
    [2] Wikipedia page about weighted arithmetic mean: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean
    [3] Wikipedia page about weighted sample variance: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance
    """
    w = 1 / unp.std_devs(a) ** 2
    avg = np.sum(w * unp.nominal_values(a), axis=axis, keepdims=True) / np.sum(w, axis=axis, keepdims=True)
    var = np.sum(w * np.subtract(avg, unp.nominal_values(a)) ** 2, axis=axis) / np.sum(w, axis=axis)
    avg = np.float64(np.squeeze(avg))
    if isinstance(avg, np.floating) and isinstance(var, np.floating):
        return ufloat(nominal_value=avg,
                      std_dev=np.sqrt(var))
    elif isinstance(avg, np.ndarray) and isinstance(var, np.ndarray):
        return unp.uarray(nominal_values=avg,
                          std_devs=np.sqrt(var))
    else:
        raise TypeError('average value and variance do not have compatible dtypes.')


def unanaverage(a, axis=None):
    """
    Computes the average and its uncertainties of numbers with uncertainties along the specified axis considering the
    uncertainties of each input value, ignoring NaNs.

    Returns the average and its uncertainty of the array elements weighted by the inverse variance of each input value
    [1]_ that results from the maximum likelihood estimator applied to aGaussian distribution. The average is calculated
    as weighted arithmetic mean [2]_ and the uncertainty is calculated as square root of the weighted sample variance of
    the input array [3]_.
    Both weighted arithmetic mean and weighted sample variance is taken over the flattened array by default, otherwise
    over the specified axis.

    Parameters
    ----------
    a : uncertainties.unumpy.ndarray
        Array containing data with uncertainties to be averaged. If `a` is not an array, a conversion is attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which to average `a`.  The default, axis=None, will average over all the elements of the
        input array.
        If axis is negative it counts from the last to the first axis.
        If axis is a tuple of ints, averaging is performed on all the axes specified in the tuple instead of a single
        axis or all the axes as before.

    Returns
    -------
    retval : uncertainties.unumpy.ndarray or uncertainties.ufloat
        Returns the averagea and its uncertainty along the specified axis weighted by the inverse variance of each input
        value.
        The result dtype follows a general pattern and is uncertainties.ufloat if `axis` is None or else

    See Also
    --------
    uaverage : weighted uncertainty average across array propagating NaNs.
    numpy.isnan : Show which elements are NaN.
    numpy.isfinite: Show which elements are not NaN or +/-inf.

    References
    ----------
    [1] Guide to the expression of uncertainty in measurement: https://www.bipm.org/utils/common/documents/jcgm/JCGM_100_2008_E.pdf
    [2] Wikipedia page about weighted arithmetic mean: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean
    [3] Wikipedia page about weighted sample variance: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance
    """
    w = 1 / unp.std_devs(a) ** 2
    avg = np.nansum(w * unp.nominal_values(a), axis=axis, keepdims=True) / np.nansum(w, axis=axis, keepdims=True)
    var = np.nansum(w * np.subtract(avg, unp.nominal_values(a)) ** 2, axis=axis) / np.nansum(w, axis=axis)
    avg = np.float64(np.squeeze(avg))
    if isinstance(avg, np.floating) and isinstance(var, np.floating):
        return ufloat(nominal_value=avg,
                      std_dev=np.sqrt(var))
    elif isinstance(avg, np.ndarray) and isinstance(var, np.ndarray):
        return unp.uarray(nominal_values=avg,
                          std_devs=np.sqrt(var))
    else:
        raise TypeError('average value and variance do not have compatible dtypes.')



# refilling_switch_on.py by JB@KIT 12/2015 jochen.braumueller@kit.edu
# client script to be run on the control machine (Raspberry Pi) to switch on nitrogen refilling, server script needs to be running
# use: only from remote as emergency

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import socket
import sys
import time
import os.path

HOST, PORT = 'ip-address', 9989

def server_request(data_string):
    
    try:
        # Create a socket (SOCK_STREAM means a TCP socket)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Connect to server and send data
        sock.connect((HOST, PORT))
        sock.sendall(data_string + "\n")
        # Receive data from the server and shut down
        received = str(sock.recv(1024))
    finally:
        sock.close()
        #print 'hallo', received
        return received

if __name__ == "__main__":

    print 'Switching on nitrogen refilling...'
    x = server_request('output on')
    s = time.ctime() + ':   deactivate manual off by separate client script - switch on'
    try:
        with open('monitor.log','a') as f: f.write(s + '\n')   #append to log
    except Exception:
        print 'Warning: No log entry written.'
    print 'Testing...'
    time.sleep(1)
    if bool(int(server_request('get_status')[0])):
        print 'Successful.'
    else:
        print 'WARNING: Switching on not successful'



# Filename: avoided_crossing_fit.py
# Alexander Stehli <alexander.stehli@kit.edu>, 05/2017
# Updates: 05/2017
# Extraction of coupling strenghts from avoided level crossings.

# Import and usage:
#    See Notebook.
# For further information see doc strings

import numpy as np
import scipy.optimize as so
import matplotlib.pyplot as plt
import inspect

class ACF():
    """
    Avoided crossing fit class (AS@KIT 2017):
    Useful for extracting the coupling strenght from avoided level crossings between an arbirary number of functions.
    
        Attributes:
            set_xdata:     Set x-data for the fit.
            set_ydata:     Set y-data for the fit.
            set_functions: Set functions of the undressed/uncoupled systems. You can also define functions by yourself.
                           Here it is important, that the function input is formatted as: New_fct(x, par1, par2, ...)
            set_init_pars: Set initial parameters for the fit (not necessary for an anticrossing between two curves).
            set_all:       Set all afore mentioned variables.

            fit:            Perform fit and generate results attribute. The latter stores the fit results and corresponding errors (least squares).
            plot_init_pars: Plot xy-input coupled modes for the initial parameters. This can be useful for finding good starting values.
            plot_results:   Plot xy-input and fit results.
            crossing_fct:   Calculates the energy levels of a coupled system of an arbirary number of modes (assuming transverse/XX-coupling of the modes).
            print_results:  Print fit results (crudely formatted).
    """    
    def __init__(self):
        self.xdata = 0
        self.ydata = 0
        self.functions = None
        self.p0 = None
        self.fit_pars = 0
        self.cov_mat = 0
        self.results = 0
    
    # Small library of inbuilt functions:      
    def constant_line(self, x, a):
        """
        A constant line.

        Args:
            x: x-value (allows for the output of arrays)
            a: return value
        
        Returns:
            An array with same length as x and value: a
        """
        return a + 0 * x
    
    def straight_line(self, x, a, b):
        """
        A straight line with slope a and offset b.

        Args:
            x: x-value
            a: Slope
            b: Offset in y-direction
        
        Returns:
            An array with same length as x and return value: a * x + b
        """
        return a * x + b
    
    def parabola(self, x, a, b, c):
        """
        A parabola with curvature a around x = b with offset c.

        Args:
            x: x value
            a: second derivative / 2
            b: Offset in x-direction
            c: Offset in y-direction

        Returns:
            An array with same length as x and return value: a*(x - b)**2 + c
        """
        return a * (x - b)**2 + c

    def hyperbola(self, x, a, b, c):
        """
        A hyperbola around x = b with offset c.

        Args:
            x: x value
            a:
            b: Offset in x-direction
            c: Offset in y-direction

        Returns:
            An array with same length as x and return value: a*(x - b)**2 + c
        """
        return np.sqrt(a * (x - b) ** 2 + c)
    
    def transmon_f01(self, x, w0, L, I_ext, djj):
        """
        Dispersion of a tunable transmon qubit with junction asymmetry.

        Args:
            x:     x-value
            w0:    Maximum qubit frequency without detuning.
            L:     Oscillation period in current/x-value.
            I_ext: Offset current/x offset.
            djj:   Josephson junction asymmetry (Ic1 - Ic2)/(Ic1 + Ic2)
        
        Returns:
            Primal transition frequency of a transmon qubit.
        """
        return  w0 * (np.abs(np.cos(np.pi/L*(x - I_ext)))*(1 + djj**2*np.tan(np.pi/L*(x - I_ext))**2)**.5)**0.5
    

    def set_xdata(self, *args):
        """
        Set x-datasets for the fit.
        x-arrays should be ordered by eigenvalue of the y-arrays (ascending order).

        Args:
            x-arrays of the undressed modes (at least 2).
            A list of arrays is also allowed as input.
        """
        self.xdata = self._list_wrapper(args)
        self._xlen = len(self.xdata)
        if self._xlen <= 1:
            print("At least 2 x-arrays required.")
        return
    
    def set_ydata(self, *args):
        """
        Set y-datasets for the fit.
        y-arrays should be ordered by eigenvalue (ascending order).

        Args:
            y-arrays of the undressed modes (at least 2).
            A list of arrays is also allowed as input.
        """
        self.ydata = self._list_wrapper(args)
        self._ylen = len(self.ydata)
        if self._ylen <=1:
            print("At least 2 y-arrays required.")
        return
    
    def set_functions(self, *args):
        """
        Set functions for the fit.
        Each function should describe one of the undressed modes (i.e. without interaction/level repelling).

        Args:
            functions of the undressed modes (at least 2). 
            A list of functions is also allowed as input.
        """
        self.functions = self._list_wrapper(args)
        self._flen = len(self.functions)
        if self._flen <= 1:
            print("At least 2 functions are required.")
            return
        # Check if functions are callable. 
        # Determine number and name of free parameters of each function.
        self._fct_par_nums = []    # List of number of each functions parameters
        self._fct_par_names = []   # List of names of each functions parameters
        for fct in self.functions:
            if callable(fct):
                sig = inspect.signature(fct)
                par_names = list(sig.parameters.keys())
                try:
                    # Remove self as fct argument if function is from acf class
                    par_names.remove("self")
                except:
                    pass
                try:
                    # Remove x as fct argument
                    par_names.remove("x")
                except:
                    print("In " + fct.__name__ + ": No x argument found. Please rename axis variable to x.")
                self._fct_par_names.append(par_names)
                self._fct_par_nums.append(int(len(par_names)))
            else:
                print("At least one function is not callable.")
                return
        # Increase number of parameters by number of coupling coefficients
        self._fct_par_nums.append(int((self._flen**2 - self._flen)/2))
        
        return
    
    def set_init_pars(self, *args):
        """
        Set initial parameters for the fit.
        Order should be: parameter1 of fct1, par2 fct1, ..., g12, g13, ...
        If the number of input parameters is insuffcient they are later on set to 1.

        Args:
            List of initial parameters.
        """
        self.p0 = self._list_wrapper(args)
        self._p0len = len(self.p0)
        return
    
    def set_all(self, x, y, f = None, p0 = None):
        """
        Set initial parameters for the fit.

        Args:
            x:  x-datasets for the fit (at least 2 arrays, can be entered as a list).
                x-arrays should be ordered by eigenvalue (ascending order).
            y:  y-datasets for the fit (at least 2 arrays, can be entered as a list).
                y-arrays should be ordered by eigenvalue (ascending order).
            f:  functions for the fit (at least 2, can be entered as a list). 
                Each function should describe one of the undressed modes (i.e. without interaction/level repelling). 
                By default, functions are set to constant_line and straight_line.
            p0: Initial parameters for the fit. Order should be: parameter1 of fct1, par2 fct1, ..., g12, g13, .. 
                If no initial parameters are given all values are set to 1.
        """
        self.set_xdata(x)
        self.set_ydata(y)
        if f != None:
            self.set_functions(f)
        elif self.functions is None:
            print("Default functions set to constant_line and straight_line.\n")
            self.set_functions(self.constant_line, self.straight_line)
        if p0 != None:
            self.set_init_pars(p0)
        
        self._validity_check()
        self._check_p0()
        return
    
    def _list_wrapper(self, args):
        """
        Wraps input for x-data, y-data, functions and init_pars in a list.

        Args:
            List or touple.
        
        Returns:
            List of input arguments.
        """
        if len(args) > 1:
            return list(args)
        elif (len(args) == 1) and isinstance(args[0], list):
            return args[0]
        else:
            print("Wrong input format.")
            return [0]
    
    def _validity_check(self):
        """
        Checks if input (x-data, y-data, functions) is valid, i.e. have the same dimension/length.

        Returns:
            Bool for validity of the input.
        """
        valid = 1
        
        # Check if parameters have been set.
        if self.xdata is 0:
            print("No xdata found. Use set_xdata for input.")
            valid = 0
        if self.ydata is 0:
            print("No ydata found. Use set_ydata for input.")
            valid = 0
        if self.functions is None:
            print("No input functions found...")
            print("constant_line and straight_line set as default.\n")
            self.set_functions(self.constant_line, self.straight_line)
        if not valid:
            return

        # Check if the three lists are of the same length.
        if self._xlen != self._ylen:
            print("Number of x- (" + str(self._xlen) + ")  and y-arrays (" + str(self._ylen) + ") is incompatible.")
            valid = 0
        if self._xlen != self._flen:
            print("Number of x-arrays (" + str(self._xlen) + ") and number of functions (" + str(self._flen) + ") is incompatible.")
            valid = 0
        if self._ylen != self._flen:
            print("Number of y-arrays (" + str(self._ylen) + ") and number of functions (" + str(self._flen) + ") is incompatible.")
            valid = 0
        if not valid:
            return valid
        
        # Check if pairs of x- and y-arrays in xdata and ydata have the same length.
        for i in range(0, self._xlen):
            if len(self.xdata[i]) != len(self.ydata[i]):
                print("In xdata/ydata: Elements " + str(i) +" do not have the same lengths (" + str(len(self.xdata[i])) + ", " + str(len(self.ydata[i])) + ").")
                valid = 0
                
        return valid
    
    def _check_p0(self):
        """
        Checks if initial parameters p0 have the correct format.
        For an insufficent number of parameters the remaining ones are set to 1.
        If p0 is too long it is truncated to the necessary length.
        """
        p0_reset = False
        # If p0 is not set, set all values to 1.
        if self.p0 is None:
            self.p0 = [1]*int(sum(self._fct_par_nums))
            p0_reset = True
            print("No initial parameters given. Setting all values to 1.")
        # If number of initial parameters is to small/large -> cut/append.
        elif self._p0len < sum(self._fct_par_nums):
            self.p0 = np.append(self.p0, [1]*int(sum(self._fct_par_nums) - self._p0len))
            p0_reset = True
            print("Not enough initial parameters. Filling up with 1.")
        elif self._p0len > sum(self._fct_par_nums):
            self.p0 = self.p0[:int(sum(self._fct_par_nums))]
            p0_reset = True
            print("To many initial parameters given. Cutting off redundant values.")
        
        if p0_reset:
            self._p0len = sum(self._fct_par_nums)
            print("New initial parameters:\n p0 = " + str(self.p0))
        return
    
    def _sort(self):
        """
        Atempts to order the pairs of x-y-arrays by eigenvalue (ascending order).
        Sorting is necessary for the fit to converge properly.
        Here this is achieved by sorting according to the mean value of the y-datasets.
        """
        # Sort ydata by mean value.
        indices = np.argsort([np.mean(i) for i in self.ydata])
        self.xdata = [self.xdata[i] for i in indices]
        self.ydata = [self.ydata[i] for i in indices]    
        if min(np.gradient(indices)) < 0:
            print("\nChanging order of x- and y-data respectively...\n")
        return
    
    def _reshape(self, pars):
        """
        Reshape pars into a list of parameters, as is accepted by crossing_fct.

        Args:
            pars: Parameters to be reshaped.
        
        Returns:
            List of parameters, the i'th element contains the fit parameters of function i.
            The last element of fit_pars contains the values for the coupling strenghts.
        """
        pars_new = []
        for fct_par_num in self._fct_par_nums:
            pars_new.append(pars[0:fct_par_num])
            pars = pars[fct_par_num:]
            
        return pars_new
    
    def _generate_results(self):
        """
        Generates a list which contains the results and errors for all parameters.
        """
        # Generate list of function indices and parameter names
        fct_ind = []
        fit_pars = np.concatenate(self.fit_pars)
        par_names = np.concatenate(self._fct_par_names)
        for i in range(len(self._fct_par_nums) - 1):
            fct_ind.append([int(i + 1)]*self._fct_par_nums[i])
        fct_ind = np.concatenate(fct_ind)
        
        for i in range(self._flen - 1):
            for j in range(i + 1, self._flen):
                fct_ind = np.append(fct_ind, (i + 1)*10 + (j + 1))
                par_names = np.append(par_names, "g_" + str(fct_ind[-1]))
        self.results = list(zip(fct_ind, par_names, fit_pars, self.std_dev))
        return
    
    def _least_square_val(self, pars):
        """
        Calculates the deviation of the coupled mode functions from the xy-input values for input parameters.
        This is deviation is minimized in the fit routine (least squares).

        Args:
            pars: Parameters from the fit.
        
        Returns:
            Deviation of coupled mode functions from xy-input.
        """
        dev = []
        pars = self._reshape(pars)
        
        for i, fct in enumerate(self.functions):
            dev = np.append(dev, self.ydata[i] - self.crossing_fct(self.xdata[i], pars)[:, i])
            
        return dev
    
    def crossing_fct(self, x, pars):
        """
        Evaluates the functions of the coupled modes (diagonalizes the matrix).

        Input:
            x:        x-values where the the functions are to be evaluated
            fct_pars: List of arrays, each array contains parameters of the corresponding (undressed) function.
                      The last array in the list contains values for the coupling strenghts.
        
        Returns:
            List of arrays, where the i'th array cointains the y-values of branch i of the coupled system.
        """
        x = np.atleast_1d(x)
        fct_pars = pars[:-1]
        g = np.atleast_1d(pars[-1])

        #Create the nondiagonal symmetric interaction matrix
        int_mat = np.zeros((self._flen, self._flen))

        for i in range(self._flen-1):
            int_mat[i, i+1:] = g[:np.size(int_mat[i,i+1:])]
            g = g[self._flen-1-i:]
        
        int_mat = int_mat+int_mat.T
        
        func_vals = np.zeros((np.size(x), self._flen))
        
        for n in range(np.size(x)):
            #Create diagonal parts of the matrix
            d_mat = np.zeros((self._flen, self._flen))
            for i, fct in enumerate(self.functions):
                d_mat[i,i] = fct(x[n], *fct_pars[i])
            
            mat = d_mat + int_mat
            func_vals[n, :] = np.linalg.eigvalsh(mat)

        return func_vals
    
    def fit(self, show_data = True, show_plot = True):
        """
        Fit functions of the coupled modes to the stored xy-data.
        Generates the results attribute, i.e. a list containing results for each free parameters:
            1. Index of the function the parameter belongs to
               For the coupling strenght this indicates the modes it belongs to
            2. Name of the parameter
            3. Value as extracted from the fit
            4. Standard deviation of the value
        
        Args:
            show_data: If true, fit data is displayed after the fit.
            show_plot: If true, plot is displayed after the fit.
        """
        # Check if fit is executable:
        if not self._validity_check():
            print("\nAddress problems before fit can be excecuted.")
            return
        
        self._check_p0()
        self._sort()
        
        fit_results = so.leastsq(self._least_square_val, self.p0, full_output = True)
        self.fit_pars, self.cov_mat = fit_results[0], fit_results[1]
        # Multiply covariance matrix with reduced chi squared to get standard deviation.
        if self.cov_mat is not None:
            self.cov_mat *= (sum(self._least_square_val(self.fit_pars)**2)/
                             (len(np.concatenate(self.xdata)) - float(len(self.p0))))
            # Calculate standard deviation
            self.std_dev = np.abs(np.diag(self.cov_mat))**0.5
        else:
            print("Covariance matrix could not be calculated.")
            self.std_dev = [float('nan')]*sum(self._fct_par_nums)
        
        # Reshape fit_pars into a list of parameters, as is accepted by crossing_fct.
        # After reshaping fit_pars is a list, where the element i contains the fit parameters of function i. 
        # The last element of fit_pars contains the coupling strenghts.
        self.fit_pars = self._reshape(self.fit_pars)
        self._generate_results()
        
        # Show output. Includes some crude formatting.
        if show_data:
            self.print_results()
                
        # Show plot.
        if show_plot:
            self.plot_results()

        return

    def print_results(self):
        """
        Prints the fit results in a legible form.
        """
        if self.fit_pars is 0:
            print("Fit was not yet performed. No data available.")
            return
        format_str = " = ( {:7.4g} +- {:7.4g})"
        fct_index = 0
        for data_set in self.results:
            if (data_set[0] > fct_index) and (data_set[0] <= self._flen):
                print("\nCurve " + str(fct_index + 1) + " (" + self.functions[fct_index].__name__ + "):")
                fct_index += 1
            elif (fct_index == self._flen) and (data_set[0] > self._flen):
                fct_index += 1
                print("\nCoupling strengths:")
            print("    " + data_set[1] + format_str.format(data_set[2], data_set[3]))
        
        return

    def _plot(self, pars):
        """
        Plots the input xy-datasets, as well as the coupled modes for the input parameters.

        Args:
            pars: Parameters of coupled modes for the plot.
        """
        dat_cols = np.array(['#0000FF', '#FF0000', '#008000', '#00CCCC', '#FF7F0E', '#CC00CC', '#000000'])
        fct_cols = np.array(['#0000FF70', '#FF000070', '#00800070', '#00CCCC50', '#FF7F0E50', '#CC00CC50', '#00000050'])
        mrkr = "*"
        if self._flen > len(dat_cols):
            dat_cols *= np.ceil(float(self._flen)/float(len(dat_cols)))
            fct_cols *= np.ceil(float(self._flen)/float(len(dat_cols)))
        
        if (self.xdata is 0) or (self.ydata is 0):
                print("No data for plot available.")
                return
        for i in range(0, self._xlen):
            plt.plot(self.xdata[i], self.ydata[i], dat_cols[i], marker = mrkr, linewidth = 0)
        xlin = np.linspace(np.nanmin(np.concatenate(self.xdata)), np.nanmax(np.concatenate(self.xdata)), 1000)
        for i in range(self._flen):
            plt.plot(xlin, self.crossing_fct(xlin, pars)[:, i], fct_cols[i])
        plt.xlim(np.nanmin(np.concatenate(self.xdata)), np.nanmax(np.concatenate(self.xdata)))
        plt.ylim(np.nanmin(np.concatenate(self.ydata)), np.nanmax(np.concatenate(self.ydata)))
        plt.show()
        return
    
    def plot_init_pars(self):
        """
        Plot of input xy-datasets and coupled modes for the initial parameters.
        This is useful to find reasonable initial values for the fit.
        """
        self._check_p0()
        self._sort()
        self._plot(self._reshape(self.p0))
        return
    
    def plot_results(self):
        """
        Plot of input xy-datasets and the coupled modes using the fit results.
        """
        if self.fit_pars is 0:
            print("Fit was not yet performed. Use self.plot_init_pars() to look at your input.")
            return
        self._plot(self.fit_pars)
        return


# This is the qtlab driver for Martinis' GHzDAC devices.
# It interfaces to a local or remote instance of the EthernetServer instrument to communicate with the pysical device.
# Markus Jerger <jerger@kit.edu>, 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# TO DO
#
#

from qkit.core.instrument_base import Instrument
import types
import logging
# comms and byte packing
import ctypes
import socket
import dpkt
import struct
import numpy
# waiting
import time

class GHzDAC(Instrument):
    '''
    This instrument controls a GHzDAC device via the EthernetServer instrument.

    Usage:
    Initialize with
    <name> = qt.instruments.create('name', 'GHzDAC', MAC=<mac>, ip_host=<ip_host>, ip_port=<ip_port>)
    <mac> = MAC address of the device, 00:01:CA:AA:00:XX, where XX is determined by dip switches on the device
    <ip_host>:<ip_port> = 127.0.0.1:413 the EthernetServer instance to connect to
    '''

    # maximum packet size
    RECV_SIZE = 2048
    # device properties
    SRAM_SIZE = 2**13

    def __init__(self, name, MAC, ip_host = '127.0.0.1', ip_port = 413, ip_timeout = 10.):
        '''
        Initializes the Oxford Instruments Kelvinox IGH Dilution Refrigerator.

        Input:
            name (string) : name of the instrument
            MAC (string) : MAC address of the target device
            ip_host:ip_port : UDP interface address of an EthernetServer 

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self.do_set_MAC(MAC)
        self._ip_host = ip_host
        self._ip_port = int(ip_port)
        self._ip_timeout = float(ip_timeout)

        # add parameters
        self.add_parameter('MAC', type=str, flags=Instrument.FLAG_GETSET)
        self.add_parameter('ip_host', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_port', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_timeout', type=float, flags=Instrument.FLAG_GETSET)
        self.add_parameter('rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_function('get_regfile')
        self.add_function('set_regfile')
        self.add_function('set_sram')
        self.add_function('get_all')
        self.add_function('seq_start')
        self.add_function('seq_stop')
        self.add_function('seq_single')
        self.add_function('seq_loop')

        # open sockets
        self._init_udp()

        # initialize hardware
        self._reg = None
        self.init_pll()
        self.init_dac(0, signed = True)
        self.init_dac(1, signed = True)
        
        # trigger register readback
        self._reg = None
        self.update_regfile()

        # update ui
        self.get_all()

    def __del__(self):
        ''' clean up '''
        pass

    def _init_udp(self):
        ''' open udp socket to EthernetServer '''
        logging.debug(__name__ + ' : Connecting to %s:%d.'%(self._ip_host, self._ip_port))
        self._udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._udp_socket.settimeout(self._ip_timeout)
        self._udp_socket.connect((self._ip_host, self._ip_port))

    def get_all(self):
        ''' call all getters to update the ui '''
        self.get_MAC()
        self.get_ip_host()
        self.get_ip_port()
        self.get_ip_timeout()
        self.get_rcvbuf()

    #
    # data structures
    #

    class sram_write:
        '''
        write-to-sram packet
        '''
        def __init__(self, addr_start, samples, markers = None):
            '''
            create new sram write packet

            Input:
                addr_start - first write address
                samples, markers - initializers; see pack_samples
            '''
            self.addr_start = addr_start
            self.samples = (256*ctypes.c_uint32).from_buffer(self.pack_samples(samples, markers).data)

        def pack_samples(self, channels, markers = None):
	        '''
	        pack sample data into 32bit ints

	        In the current hardware, each analog channel has a resolution of 14bits 
	        and each marker has a resolution of one bit.

	        Input:
		        channels - 2xN array of integers
		        markers  - 4xN array of bools
	        Output:
		        N array of packet samples
	        '''
	        if len(channels) != 2:
		        print 'pack_samples requires exactly 2 analog channels'
	        if (markers != None) and (len(markers) != 4):
		        print 'pack_samples requires exactly 4 marker channels'
	        channels = numpy.array(channels, numpy.uint32)
	        samples =  (channels[0,:] & 0x00003FFF)
	        samples |= (channels[1,:] & 0x00003FFF) << 14
	        if(markers != None):
		        markers = numpy.array(markers, numpy.uint32)
		        samples |= (1<<31)*markers[0] + (1<<30)*markers[1] + (1<<29)*markers[2] + (1<<28)*markers[3]
	        return samples

        def pack(self):
            ''' return string reprentation of this sram write packet '''
            if(len(self.samples) != 256):
                logging.warning(__name__ + ' : number of samples per packet must be 256.')
            return struct.pack('<H256I', self.addr_start, *self.samples)

    class regfile:
        '''
        GHZdac register file

        r/w:
            sequencer - SEQ_* sequencer mode
            readback - REG_* readback mode
            i2c_stop, i2c_rw, i2c_ack_tx - unknown I2C stuff
            i2c_data_tx - unknown array of 7 I2C data bytes
            registers - array of 30 register bytes
            slave - daisy-chain master/slave select
            delay - start delay, in ns?, 8bit
            sync - master start delay, in 4ns clock cycles
            daclk - DACLK_* d/a converter clocking flags
            serial_tx - unknown, 32 bits
        r/o:
            build - FPGA code build number
            memsum - memory checksum (all zeros)
            serial_rx - byte received from serial DAC interface
            clkmon - clock status
        '''
        # constants
        # register file packet size
        SIZE_TX = 56+14
        SIZE_RX = 70+14

        # sequencer starting mode
        SEQ_STOP = 0x0
        SEQ_MEMORY = 0x1
        SEQ_TEST = 0x2
        SEQ_LOOP = 0x3 # stop with SEQ_SINGLE or SEQ_STOP
        SEQ_SINGLE = 0x4
        SEQ_PAGE0 = 0x00
        SEQ_PAGE1 = 0x80

        # register readback mode
        REG_NOREADBACK = 0x00 	# no readback
        REG_READDELAY = 0x01 	# readback after 2us
        REG_READI2C = 0x02	# readback synchronized to I2C
        REG_READSTREAM = 0x03	# continuous readback (stream timing data)

        # D/A converter clocking
        DACLK_A_EN = 0x10
        DACLK_B_EN = 0x20
        DACLK_A_INV = 0x01
        DACLK_B_INV = 0x02
        DACLK_RESET = 0x80

        # r/w registers
        sequencer = SEQ_STOP
        readback = REG_NOREADBACK
        i2c_stop = 0x0
        i2c_rw = 0x0
        i2c_ack_tx = 0x0
        i2c_data_tx = numpy.zeros(8, numpy.uint8)
        registers = numpy.zeros(30, numpy.uint8)
        slave = False
        delay = 0
        sync = 0
        daclk = DACLK_A_EN | DACLK_B_EN
        serial_tx = 0x00000000
        # r/o registers
        build = None
        memsum = None
        sermon = None
        serial_rx = None
        clkmon = None
        i2c_ack_rx = None
        i2c_data_rx = None
	
        def __init__(self, **kwargs):
	        '''
	        initialize object, copying keyword args into properties of the same name
	        '''
	        for key, value in kwargs.iteritems():
		        if hasattr(self, key):
			        setattr(self, key, value)
		        else:
			        print 'reg: unsupported property %s'%key

        def pack(self):	
	        '''
	        update device registers

	        Output:
		        ethernet packet type number and payload string
	        '''
	        i2c_data_tx = str(numpy.array(self.i2c_data_tx, numpy.uint8).data)
	        registers = str(numpy.array(self.registers, numpy.uint8).data)
	        slave = numpy.uint8(self.slave)
	        length = 56
	        payload = struct.pack(
		        '<BB BBB8s 30s BBBB i xxxxx', 
		        self.sequencer, self.readback, 
		        self.i2c_stop, self.i2c_rw, self.i2c_ack_tx, i2c_data_tx, 
		        registers, 
		        slave, self.delay, self.sync, self.daclk,
		        self.serial_tx
	        )
	        return (length, payload)

        def unpack(self, payload):
	        '''
	        parse register file sent by device

	        Input:
		        payload of a register readback packet sent by the device
	        '''
	        (
		        self.sequencer, self.readback, 
		        self.i2c_stop, self.ic2c_rw, self.i2c_ack_tx, i2c_data_tx, 
		        registers,
		        slave, self.delay, self.sync, self.daclk, 
		        self.serial_tx,
		        self.build, self.memsum, self.serial_rx, self.sermon, self.clkmon,
		        self.i2c_ack_rx, i2c_data_rx
	        ) = struct.unpack( '<BB BBB8s 30s BBBB I BIBBB xx B8s', payload)
	        self.registers = numpy.fromstring(registers, numpy.uint8)
	        self.i2c_data_tx = numpy.fromstring(i2c_data_tx, numpy.uint8)
	        self.i2c_data_rx =  numpy.fromstring(i2c_data_rx, numpy.uint8)

        def __str__(self):
	        '''
	        return string representation of self
	        '''
	        props = []
	        for prop in dir(self): 
		        if not callable(getattr(self, prop)):
			        props += ['%s: %s'%(prop, getattr(self, prop))]
	        return '\n'.join(props)

    #
    # communication-related parameters
    #

    def do_set_MAC(self, MACin):
        '''
            convert mac address to byte string
            expected input format is 01:23:45:67:89:AB
        '''
        # split input and convert the parts to ints
        MAC = [int(byte, base = 16) for byte in str(MACin).split(':')]
        # merge the parts into a byte string
        self._MAC = struct.pack('>6B', *MAC)

    def do_get_MAC(self):
        ''' return device MAC address '''
        return '%02x:%02x:%02x:%02x:%02x:%02x'%(struct.unpack('>6B', self._MAC))
 
    def do_get_ip_host(self):
        ''' return hostname of the udp socket '''
        return self._ip_host

    def do_get_ip_port(self):
        ''' return port number of the udp socket '''
        return self._ip_port

    def do_set_ip_timeout(self, timeout):
        ''' set timeout for socket operations '''
        self._ip_timeout = timeout
        self._udp_socket.settimeout(timeout)

    def do_get_ip_timeout(self):
        ''' get timeout for socket operations '''
        return self._udp_socket.gettimeout()

    def do_set_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the raw socket '''
        self._udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_rcvbuf(self):
        ''' get the receive buffer size of the raw socket '''
        return self._udp_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)

    #
    # device initialization
    #
    def SPI(self, sel, sequence):
        '''
        send a sequence of words over the SPI bus
        the PLL and DACs are connected to SPI

        Input:
            sel - hardware to send to (1: PLL, 2: DAC A, 3: DAC B)
            sequence - sequence of 24bit words to transmit
        '''
        reg = self._reg if (self._reg != None) else self.regfile()
        result = []
        for data in sequence:
            reg.serial_tx = (data << 8) | sel
            self.update_regfile(reg)
            result += [self._reg.serial_rx]
            time.sleep(0.01)
        reg.serial_tx = 0
        return result


    def init_pll(self):
        '''
        program AD4107 PLL to generate 1GHz output clock for the DACs and digital outputs

        setup:
	        prescaler P=8
	        clk/ref ratio = 100 = (P*B+A)/R
	        choosing R=1, B=12, A=4
        latches:
	        24bit values, last two bits are latch selection (C2, C1)
	        (1, 1) initialize  = 	0x1fc093
	        (1, 0) function = 	0x1fc092 (negative pd polarity, output lock detect on mux, high current out)
	        (0, 0) R = 		0x000004 (14bit R)
	        (0, 1) N = 		0x000C11 (13bit B, 3..8191), (6 bit A), high precision
        sequence:
	        initialize, function, R, N

        Output:
	        an array of 32bit values to place into GHZdac serial registers
        '''
        #return [0x00008301, 0x00008201, 0x10000401, 0x000C1101, 0x00009201]
        sel = 0x01 # PLL
        seq = [0x1FC093, 0x1FC092, 0x100004, 0x000C11]
        self.SPI(sel, seq)

    def init_dac(self, channel, signed = True):
        '''
        program ADxxxx DAC

        Input:
            channel number - integer (0 or 1)
            signed - dac input signed/unsigned flag
        '''
        seq = [0x000024, 0x000004, 0x001603, 0x000500] if signed else \
              [0x000026, 0x000006, 0x001603, 0x000500] # 0x050000 
        sel = 0x02 + channel
        self.SPI(sel, seq)

    def init_lvds(self, channel):
        '''
            calibrate LVDS receivers in the DACS to the center of the data eye
            the calibration procedure is found in the AD9736 data sheet
            and Martinis' control software

        Input:
            channel number - 0 or 1
        '''
        op = 2+channel # SPI operation code for DAC operation

        # vary delay until we hit a clock edge
        seq = lambda t: [0x000500 + 0x10*t, 0x008700] # adjust lvds delay
        # change of status indicates that we cross a clock edge
        t = 0
        stat0 = self.SPI(op, seq(t))[1]
        while True:
            t += 1
            if(t == 16):
                logging.error(__name__ + ' : failed to find DAC clock edge')
                return False
            stat1 = self.SPI(op, seq(t))[1]
            if(stat0 != stat1): break
        logging.debug(__name__ +  ' : clock edge at %d.'%t)
        self.SPI(op, seq((t+1) % 16))
        time.sleep(0.05)

        # adjust fifo counter to half-full
        seq = [0x0700, 0x8700, 0x0701, 0x8700, 0x0702, 0x8700, 0x0703, 0x8700]
        stat = self.SPI(op, seq)
        stat = [(stat[i] >> 4) & 0x0f for i in [1, 3, 5, 7]]
        # optimum read counter value is 3
        delay = numpy.argmin(numpy.abs(numpy.array(stat)-3))
        seq = lambda t: [0x0700 + 0x10*t, 0x8700]
        self.SPI(op, seq(delay))
        if(stat[delay] != 3):
            logging.error(__name__ + ' : failed to set fifo read counter to 3.')
            return False
        return True

    #
    # device configuration
    #
    def update_regfile(self, reg = None):
        '''
        transmit the current or an empty register file to the device and request register readback
        '''
        if(reg == None): reg = self._reg if (self._reg != None) else self.regfile()
        reg.readback = self.regfile.REG_READDELAY
        self.set_regfile(reg)

    def get_regfile(self):
        '''
        return buffered register file
        '''
        return self._reg

    def set_regfile(self, reg):
        '''
        transmit a register file packet to the device
        '''
        self._reg = reg
        length, payload = reg.pack()
        packet = dpkt.ethernet.Ethernet(dst = self._MAC, type = length, data = payload)
        self._udp_socket.send(packet.pack())
        if(reg.readback == self.regfile.REG_READDELAY):
            reg = self.read_regfile()
            if(reg != None): self._reg = reg

    def read_regfile(self):
        '''
        wait for a register readback packet
        '''
        try:
            packet_raw = self._udp_socket.recv(self.RECV_SIZE)
        except Exception as e:
            logging.error(__name__ + ' : register readback failed. reason: %s.'%(e))
            return None
        # basic sanity checks
        if(len(packet_raw) != self.regfile.SIZE_RX):
            logging.error(__name__ + ' : register readback from %s has unexpected length %d.'%(self.do_get_MAC(), len(packet_raw)))
            return None
        packet = dpkt.ethernet.Ethernet(packet_raw)
        if(packet.src != self._MAC):
            logging.error(__name__ + ' : register readback MAC address does not match.')
            return None
        # parse register file
        reg = self.regfile()
        reg.unpack(packet.data)
        return reg

    def set_sram(self, addr_start, samples, markers = None):
        '''
        send write-to-sram packet to the device 
        '''
        if(addr_start >= self.SRAM_SIZE):
            print 'sram_write: start address must be smaller than sram memory size'
        payload = self.sram_write(addr_start, samples, markers).pack()
        packet = dpkt.ethernet.Ethernet(dst = self._MAC, type = len(payload), data = payload)
        self._udp_socket.send(packet.pack())

    def seq_start(self):
        self._reg.sequencer = self._reg.SEQ_START
        self.update_regfile()
    def seq_stop(self):
        self._reg.sequencer = self._reg.SEQ_STOP
        self.update_regfile()
    def seq_single(self):
        self._reg.sequencer = self._reg.SEQ_SINGLE
        self.update_regfile()
    def seq_loop(self):
        self._reg.sequencer = self._reg.SEQ_LOOP
        self.update_regfile()



# virtual_MultiplexingReadout.py
# initiated and written by M. Jerger 2011/2012
# updated by A. Schneider <andre.schneider@kit.edu> and J. Braumueller <jochen.braumueller@kit.edu>, 05/2016, 08/2016, 04/2018

# This script may be regarded as a wrapper, handling data acquisition with the ADC (commonly known here as mspec)
# and readout pulse generation by the DAC (which is the AWG used for generating readout pulses).

# Now, in 2016, the Martinis AWGs are outdated (and still buggy) as they do not provide longer readout pulses
# due to a limitation in memory.
# The script is updated to also be usable with other AWGs (Tabor, Tektronix).

"""
usage:

readout = qt.instruments.create('readout','virtual_MultiplexingReadout',sample=qubit)
            
Mandatory sample attributes:
    readout_awg, readout_mw_src, mspec
"""

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from qkit.core.instrument_base import Instrument
import types
import logging
import numpy as np
import scipy.special
from scipy import signal
from time import sleep
from qkit.measure.timedomain.awg import load_awg as lawg


class virtual_MultiplexingReadout(Instrument):

    def __init__(self, name, sample):
        # ToDo: object should not set params previously set by the user

        Instrument.__init__(self, name, tags=['virtual'])

        #self._instruments = instruments.get_instruments()

        # Add parameters
        self.add_parameter('LO',    type=float, flags=Instrument.FLAG_GETSET, units='Hz')
        self.add_parameter('tone_freq', type=list, flags=Instrument.FLAG_GETSET, units='Hz')
        self.add_parameter('tone_relamp', type=list, flags=Instrument.FLAG_GETSET, units='')
        self.add_parameter('tone_pha', type=list, flags=Instrument.FLAG_GETSET, units='rad')
        self.add_parameter('tone_amp', type=float, flags=Instrument.FLAG_GETSET, units='V')
        self.add_parameter('adc_clock',     type=float, flags=Instrument.FLAG_GET, units='Hz')
        self.add_parameter('dac_clock',     type=float, flags=Instrument.FLAG_SET, units='Hz')
        self.add_parameter('dac_duration',  type=float, flags=Instrument.FLAG_GETSET, units='s')
        self.add_parameter('dac_channel_I', type=int, flags=Instrument.FLAG_SET)
        self.add_parameter('dac_channel_Q', type=int, flags=Instrument.FLAG_SET)
        self.add_parameter('adc_channel_I', type=int, flags=Instrument.FLAG_SET)
        self.add_parameter('adc_channel_Q', type=int, flags=Instrument.FLAG_SET)
        self.add_parameter('dac_attack', type=float, flags=Instrument.FLAG_SET)
        self.add_parameter('dac_decay', type=float, flags=Instrument.FLAG_SET)
        self.add_parameter('global_pha', type=float, flags=Instrument.FLAG_SET)
        self.add_parameter('dac_delay', type = float, flags=Instrument.FLAG_GETSET)

        self.add_function('update')
        self.add_function('readout')

        self.sample = sample
        self._mspec = sample.mspec
        self._tone_relamp = None
        self._tone_pha = None
        self._LO = 0
        self._tone_amp = 1
        
        try:
            self._awg = sample.readout_awg
        except AttributeError:
            self._awg = None
            logging.warning("You haven't provided a readout_awg. I can't set the readout pulse for you."
                            "Please also think of specifying the readout if-frequency.")
        # check that awg name exists
        if self._awg:
            try:
                self._awg.get_type()
            except AttributeError:
                logging.error('Cannot resolve awg name. Provide name attribute in awg instrument driver.')
            if "GHzDAC" == self._awg.get_type():   #Martinis board
                self.do_set_dac_delay(0)
                #self._dac_duration = 10e-9
            elif "Tabor_WX1284C" == self._awg.get_type():   #Tabor AWG
                self._awg.set_trigger_impedance(50)   #50 Ohms
                self._awg.preset_readout()   #sets runmode = 'AUTO', trigger_mode='TRIG', starts with the end of the manipulation signal
                self.do_set_dac_delay(-1)
            elif "Tektronix" in self._awg.get_type():
                self.do_set_dac_delay(0)
                self._awg.set_ch1_status(True)
                self._awg.set_ch2_status(True)
                self._awg.run()
                self._awg.wait(10,False)
            else:
                logging.error('Specified AWG unknown. Aborting.')
                raise ImportError
            #default settings
            
            try:
                self._awg.set_clock(self.sample.readout_clock)
                self._dac_clock = self._awg.get_clock()
                self._dac_channel_I = 0
                self._dac_channel_Q = 1
                self._dac_attack = 0#5e-9
                self._dac_decay = 0#5e-9
                self._phase = 0
                self._tone_freq = [30e6]
                self._dac_duration = 400e-9
                self.dac_delay = 0
            except Exception as m:
                logging.error('Defaults not set properly: '+str(m))
                raise ImportError
        self._adc_channel_I = 1
        self._adc_channel_Q = 0
        self._phase = 0
        
        # used for DDC
        self.lowpass_order = 20
        self.cut_off_freq_ratio = 0.8  # ratio of the IQ frequency up to that is transmitted
        # lowpass_delay = (lowpass_order / 2) / freqs
        # a lowpass of order N delays the signal by N/2 samples

    def get_all(self):
        self.get_LO()

    def do_get_adc_clock(self):
        """
            sample rate of the analog to digital converter
        """
        return self._mspec.get_clock()

    def do_set_dac_clock(self, clock):
        """
            sample rate of the digital to analog converter
        """
        self._dac_clock = clock

    def do_set_dac_duration(self, duration):
        """
            defines the time duration of the signal output on the dac
            this defines the readout pulse length and maximum repetition rate
        """
        self._dac_duration = duration

    def do_get_dac_duration(self):
        return self._dac_duration
        
    def do_set_dac_channel_I(self, channel):
        """
            set dac channel that outputs the I signal, -1=None
        """
        self._dac_channel_I = channel
        if(self._dac_channel_I == -1) and (self._dac_channel_Q == -1):
            logging.warning(__name__ + 'no output signal is produced if both dac_channel_I and dac_channel_Q are none')
    
    def do_set_dac_channel_Q(self, channel):
        """
            set dac channel that outputs the Q signal, -1=None
        """
        self._dac_channel_Q = channel
        if(self._dac_channel_I == -1) and (self._dac_channel_Q == -1):
            logging.warning(__name__ + 'no output signal is produced if both dac_channel_I and dac_channel_Q are none')
    
    def do_set_adc_channel_I(self, channel):
        """
            set adc channel that acquires the I signal, -1=None
        """
        self._adc_channel_I = channel
        if(self._adc_channel_I == -1) and (self._adc_channel_Q == -1):
            logging.warning(__name__ + 'no signal is acquired if both dac_channel_I and dac_channel_Q are none')
    def do_set_adc_channel_Q(self, channel):
        """
            set adc channel that acquires the Q signal, -1=None
        """
        self._adc_channel_Q = channel
        if(self._adc_channel_I == -1) and (self._adc_channel_Q == -1):
            logging.warning(__name__ + 'no signal is acquired if -1dac_channel_I and dac_channel_Q are none')

    def do_set_dac_attack(self, attack):
        """ set attack time of the readout pulse """
        self._dac_attack = attack

    def do_set_dac_decay(self, decay):
        """ set decay time of the readout pulse """
        self._dac_decay = decay

    def do_set_LO(self, frequency):
        """
        Sets the readout microwave source to a given frequency and stores this value for future evalution
        :param frequency: frequency of the readout microwave source
        :return: 
        """
        self.sample.readout_mw_src.set_frequency(frequency)
        self._LO = frequency

    def do_get_LO(self):
        return self._LO

    def do_set_tone_amp(self, amp):
        self._tone_amp = amp

    def do_get_tone_amp(self):
        return self._tone_amp

    def do_set_tone_freq(self, freqs):
        self._tone_freq = np.array(freqs)

    def do_get_tone_freq(self):
        return self._tone_freq

    def do_set_tone_relamp(self, amps):
        self._tone_relamp = np.array(amps)

    def do_get_tone_relamp(self):
        return self._tone_relamp

    def do_set_tone_pha(self, phases):
        self._tone_pha = np.array(phases)

    def do_get_tone_pha(self):
        return self._tone_pha

    def do_set_global_pha(self, phase):
        self._phase = float(phase)

    def do_get_global_pha(self):
        return self._phase
        
    def do_set_dac_delay(self,delay):
        """
        sets the pause time before the dac pulse in s
        a value <= -1 will take sample.exc_T
        """
        self.dac_delay = delay
    def do_get_dac_delay(self):
        return self.dac_delay
        


    # +++++ ADC acquisition ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def _acquire_IQ(self):
        """
        essentially do a mspec.acquire and return I and Q
        """
        data = self._mspec.acquire().swapaxes(0, 1)
        if self._adc_channel_I == -1:
            I = np.zeros(data.shape[1:])
        else:
            I = data[self._adc_channel_I]
        if self._adc_channel_Q == -1:
            Q = np.zeros(data.shape[1:])
        else:
            Q = data[self._adc_channel_Q]
        return I, Q

    def readout(self, timeTrace=False, ddc=None):
        """
        measure transmission of the tones set
            --> return amplitude and phase data at the given IQ frequency and also the full I and Q time trace
            if set to true using IQ_decode
        :param timeTrace: also output raw trace for further processing
        :param ddc: performs a digital down conversion of your readout tone to get the time trace at the frequency of interest
        :return:
        """
        Is, Qs = self._acquire_IQ()
        if ddc is None:
            if len(Is.shape) == 2:
                sig_amp = np.zeros((Is.shape[1], len(self._tone_freq)))
                sig_pha = np.zeros((Is.shape[1], len(self._tone_freq)))
                for idx in range(Is.shape[1]):
                    sig_amp[idx, :], sig_pha[idx, :] = self.IQ_decode(Is[:, idx], Qs[:, idx])
            else:
                sig_amp, sig_pha = self.IQ_decode(Is, Qs)
        else:
            if len(Is.shape) == 2:
                sig_amp = np.zeros((Is.shape[1],Is.shape[0], len(self._tone_freq)))
                sig_pha = np.zeros((Is.shape[1],Is.shape[0], len(self._tone_freq)))
                for idx in range(Is.shape[1]):
                    sig_amp[idx,:, :], sig_pha[idx,:, :] = self.digital_down_conversion(Is[:, idx], Qs[:, idx])
            else:
                sig_amp, sig_pha = self.digital_down_conversion(Is, Qs)
        if timeTrace:
            return sig_amp, sig_pha, Is, Qs
        else:
            return sig_amp, sig_pha


    def spectrum(self, segment = 0):
        """
            measure transmission of the tones set
        """
        I, Q = self._acquire_IQ()
        if(len(I.shape) == 2):
            # in segmented mode, take only one segment
            return self.IQ_fft(I[:, segment], Q[:, segment])
        else:
            return self.IQ_fft(I, Q)


    def IQ_fft(self, I, Q, samplerate = None):
        """
            calculate fourier transform of the acquired waveform

            Input:
                I, Q       - signal acquired at rate samplerate
                samplerate - rate at which I and Q were sampled
        """
        if(samplerate is None): samplerate = self.get_adc_clock()

        sig_t = np.array(I) + 1j*np.array(Q)
        sig_f = np.fft.fftshift(np.fft.fft(sig_t))
        sig_x = self._LO+np.fft.fftshift(np.fft.fftfreq(sig_t.size, 1./samplerate))

        return sig_x, np.abs(sig_f), np.angle(sig_f)


    def IQ_decode(self, I, Q, freqs = None, samplerate = None, phase = None):
        """
            calculate fourier analysis of the acquired waveform and
            return amplitude and phase of requested frequency components

            Input:
                I, Q       - signal acquired at rate samplerate
                freqs      - interesting frequency components
                samplerate - rate at which I and Q were sampled
                phase      - apply additional rotation to I+1j*Q

            Output:
                two vectors: amplitude and phase of each fft point
        """
        if samplerate is None: samplerate = self.get_adc_clock()
        if freqs is None: freqs = self._tone_freq
        if phase is None: phase = self._phase
        freqs = np.array(freqs)-self._LO

        sig_t = (np.array(I) + 1j*np.array(Q))*np.exp(1j*phase)
        return self.fourieranalysis(sig_t, freqs, samplerate)

    def fourieranalysis(self, signal_t, freqs, samplerate):
        """
        useful for only a few samples and freqs because no interpolation is needed
        :param signal_t: The complex waveform to be analyzed
        :param freqs: Float or array of Floats of frequencies
        :param samplerate:
        :return: [amplitudes, phases], each of them being an array over len(freqs)
        """
        freqs = np.atleast_1d(freqs)
        w = np.array([np.exp(-2 * np.pi * 1j * f * np.arange(len(signal_t)) / samplerate) for f in freqs])
        f_signal = np.sum(w * signal_t, axis=1) / len(signal_t)
        sig_amp = np.abs(f_signal)
        sig_pha = np.angle(f_signal)
        return sig_amp, sig_pha

    def digital_down_conversion(self, I, Q, freqs=None):
        """
        performs a digital down conversion to get rid of the carrier frequency.
        Useful for timetrace readout, when only envelope is needed.
        :param I:
        :param Q:
        :param freqs:
        :return:
        """
        if freqs is None:
            freqs = np.array(self._tone_freq) - self._LO
        samplerate = self.get_adc_clock()
        sig_amp = np.zeros((len(freqs), len(I)))
        sig_pha = np.zeros((len(freqs), len(I)))
        t = np.linspace(0, float(len(I)) / samplerate, len(I))
        for i, f in enumerate(freqs):
            cut_off_freq = self.cut_off_freq_ratio * np.abs(f) / (samplerate / 2)
            b, a = signal.butter(self.lowpass_order, cut_off_freq, 'low')  # design the filter
            signal_down = (I+1j*Q)*np.exp(1j*2*np.pi*f*t)
            signal_down_lp = scipy.signal.lfilter(b, a, signal_down)
            sig_amp[i, :] = np.abs(signal_down_lp)
            sig_pha[i, :] = np.angle(signal_down_lp)
        return sig_amp.T, sig_pha.T  # transform because readout expects the data this way

    # +++++ DAC (AWG) settings ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        
    def update(self):
        """
        General method to update all (DAC) settings
        
        create DAC waveforms and update DAC, set DAC (AWG) to run mode (and switch outputs on)
        using IQ.encode and the hidden method self._update_dac()
        """
        # calculate required IF frequencies
        tones = np.array(self._tone_freq)
        ntones = len(tones)
        IFtones = tones-self._LO

        # perform sanity check
        IFMax = max(abs(IFtones))
        # print('IF tones are ', (IFtones/1e6), 'MHz')
        # this is only a quick check as the maximum bandwidth of an ADC card is lower than half of the sampling rate
        if IFMax > self._awg.get_clock()/2:
            logging.warning(__name__ + ' : maximum IF frequency of %fMHz is above the limit of the DAC.'%(IFMax/1e6))
        if IFMax > self._mspec.get_clock()/2:
            logging.warning(__name__ + ' : maximum IF frequency of %fMHz is above the limit of the ADC.'%(IFMax/1e6))

        # build I and Q waveforms and send them to the DAC
        if self._tone_relamp is None:
            amplitudes = 1./ntones*np.ones(ntones)
        elif len(self._tone_relamp) == 1:
            amplitudes = 1./ntones*self._tone_relamp * np.ones(ntones)
        elif len(self._tone_relamp) == ntones:
            amplitudes = 1./ntones*self._tone_relamp
        else:
            print("tone_relamp shape does not fit number of multiplexed frequencies. Setting tone_relamp to 1")
            amplitudes = 1./ntones*np.ones(ntones)
        if self._tone_pha is None:
            phases = np.zeros(ntones)
        elif len(self._tone_pha) == 1:
            phases = self._tone_pha * np.ones(ntones)
        elif len(self._tone_pha) == ntones:
            phases = self._tone_pha
        else:
            phases = np.zeros(ntones)
        I, Q,m1 = self.IQ_encode(self._dac_duration, IFtones, [amplitudes, amplitudes], [phases, phases], self._dac_clock, self._dac_attack, self._dac_decay)
        self._update_dac([self._dac_channel_I, self._dac_channel_Q], [I, Q],marker1 = [m1,m1],marker2 = [m1,m1])
      
    def _update_dac(self, channels, samples, marker1 = None, marker2 = None, drive = None, path = None):
        """
            update waveform data on the awg

            Input:
                samples - matrix of samples to put on awg channels
                channels - channels to install them on
        """
        if self._awg is None:
            logging.warning("You haven't provided a readout awg so I can't load a pulse in. "
                            "You have to do it manually!")
            return
        samples = np.array(samples)
        if marker1 is None:
            marker1 = np.zeros(samples.shape)
            marker1[:,1:10]=1
        if marker2 is None:
            marker2 = np.zeros(samples.shape)
            marker2[:,1:10]=1

        if "GHzDAC" == self._awg.get_type():   #Martinis board
            drive = 'c:'
            path = '\\waveforms'
            for idx in range(0, len(channels)):
                if(channels[idx] == -1): continue # only apply I or Q if the user wishes that
                channel = 1+channels[idx]
                
                fn = drive + path + '\\ch%d'%channel
                self._awg.send_waveform(samples[idx, :], marker1[idx, :], marker2[idx, :], fn, self._dac_clock)
                self._awg.load_waveform(channel, 'ch%d'%channel, drive, path)
                
            # setup channel overall amplitude
            for channel in [self._dac_channel_I, self._dac_channel_Q]:
                getattr(self._awg, 'set_ch%d_amplitude'%(channel+1))(self._tone_amp)
            self._awg.run()
            
        elif "Tabor_WX1284C" == self._awg.get_type():   #Tabor AWG
            #self._awg.preset_readout()
            try:
                self._awg.set_clock(self.sample.readout_clock)
            except:
                logging.warning('Clock and amplitude settings not written.')
            self._awg.wfm_send2(samples[0],samples[1],m1 = marker1[0],m2 = marker2[0],seg=1)
            self._awg.define_sequence(1,1)
        elif "Tektronix" in self._awg.get_type():   #Tektronix AWG
            try:
                self._awg.set_clock(self.sample.readout_clock)
                self._awg.set_ch1_amplitude(self._tone_amp)
                self._awg.set_ch2_amplitude(self._tone_amp)
            except:
                logging.warning('Clock and amplitude settings not written.')
            if self.sample:
                lawg.update_sequence([1],lambda t, sample: [samples[channels[0]],samples[channels[1]]], self.sample, awg = self._awg, marker = [[marker1,marker2],[marker1,marker2]], show_progress_bar=False)
            else:
                logging.error('Please provide sample object in instantiating readout when using Tektronix AWG for readout.')

    def IQ_encode(self, duration, frequencies, amplitudes = None, phases = None, samplerate = None, attack = 2e-9, decay = 2e-9):
        """
            provide I and Q data that will create a single side band multitone signal
            --> generate sin and cos waveforms for I and Q with IF frequency

            Input:
                amplitudes   - normalized amplitudes (1xN) or I and Q amplitudes (2xN) of each tone
                frequencies  - IF frequencies desired
                duration     - duration of the waveform in real time
                samplerate   - sample rate of the output device
                attack/decay - with of trailing and leading edges

            Output:
                two vectors containing samples for I and Q
        """

        if(samplerate is None): samplerate = self._dac_clock
        nSamples = int(samplerate*duration)
        indices = np.arange(0, nSamples)
        frequencies = np.array(frequencies)
        omegas = 2.*np.pi*frequencies/samplerate

        if(amplitudes is None):
            amplitudes = np.ones((2, frequencies.size))
        else:
            amplitudes = np.array(amplitudes)
            if (amplitudes.ndim == 1): amplitudes = np.array([amplitudes, amplitudes])
        if(phases is None):
            phases = np.zeros((2, frequencies.size))
        else:
            phases = np.array(phases)
            if (phases.ndim == 1): phases = np.array([phases, phases])
            
        # generate multitone waveform
        I = np.zeros(int(np.ceil((nSamples+1)/16.)*16))
        Q = np.zeros(int(np.ceil((nSamples+1)/16.)*16))
        for idx in range(0, frequencies.size):
            I[1:nSamples+1] += amplitudes[0, idx]*np.sin(omegas[idx]*indices+phases[0, idx])
            Q[1:nSamples+1] += amplitudes[1, idx]*np.cos(omegas[idx]*indices+phases[1, idx])
            
        # apply envelope: erf, erf(\pm 2) is almost 0/1, erf(\pm 1) is ~15/85%
        nAttack = int(2*samplerate*attack)
        sAttack = 0.5*(1+scipy.special.erf(np.linspace(-2, 2, nAttack)))
        I[0:nAttack] *= sAttack
        Q[0:nAttack] *= sAttack
        nDecay = int(2*samplerate*decay)
        sDecay = 0.5*(1+scipy.special.erf(np.linspace(2, -2, nDecay)))
        I[(nSamples-nDecay):nSamples] *= sDecay 
        Q[(nSamples-nDecay):nSamples] *= sDecay 
        m1 = np.zeros(len(I),dtype=np.int8)
        m1[1:len(I)-1] = 1
        if self.dac_delay != 0:
            if self.dac_delay <= -1:
                dac_delay = self.sample.exc_T - self.sample.overlap + self.sample.readout_delay
            else:
                dac_delay = self.dac_delay
            I = np.append(np.zeros(int(dac_delay*samplerate/16)*16),I)
            Q = np.append(np.zeros(int(dac_delay*samplerate/16)*16),Q)
            m1 = np.append(np.zeros(int(dac_delay*samplerate/16)*16),m1)
        return I, Q,m1



class errors():
    # ***********************************************************************
    #
    # SpcErr.h = (c) Spectrum GmbH, 2006
    #
    # ***********************************************************************
    #
    # error codes of the Spectrum drivers. Until may 2004 this file was
    # errors.h. Name has been changed because errors.h has been already in
    # use by windows.
    #
    # ***********************************************************************

    ERR_OK = 0x0000  # 0 No Error
    ERR_INIT = 0x0001  # 1 Initialisation error
#    -ERR_NR = 0x0002  # 2 Board number out of range
    ERR_TYP = 0x0003  # 3 Unknown board Typ
    ERR_FNCNOTSUPPORTED =0x0004  # 4 This function is not supported by the hardware
    ERR_BRDREMAP = 0x0005  # 5 The Board Index Remap table is wrong
    ERR_KERNELVERSION = 0x0006  # 6 The kernel version and the dll version are mismatching
    ERR_HWDRVVERSION = 0x0007  # 7 The driver version doesn't match the minimum requirements of the board
    ERR_ADRRANGE = 0x0008  # 8 The address range is disabled (fatal error)
    ERR_INVALIDHANDLE = 0x0009  # 9 Handle not valid
    ERR_BOARDNOTFOUND = 0x000A  # 10 Card mit given name hasn't been found
    ERR_BOARDINUSE = 0x000B  # 11 Card mit given name is already in use by another application
# ERR_EXPHW64BITADDR = 0x000C
# ERR_FWVERSION = 0x000D
    ERR_LASTERR = 0x0010  # 16 Old Error waiting to be read
# ERR_BOARDINUSE = 0x0011
    ERR_ABORT = 0x0020  # 32 Abort of wait function
    ERR_BOARDLOCKED = 0x0030  # 48 Board acess already locked by another process. it's not possible to acess one board through multiple processes

    ERR_REG = 0x0100  #256 unknown Register for this Board
    ERR_VALUE = 0x0101  #257 Not a possible value in this state
    ERR_FEATURE = 0x0102  #258 Feature of the board not installed
    ERR_SEQUENCE = 0x0103  #259 Channel sequence not allowed
    ERR_READABORT = 0x0104  #260 Read not allowed after abort
    ERR_NOACCESS = 0x0105  #261 Access to this register denied
    ERR_POWERDOWN = 0x0106  #262 not allowed in Powerdown mode
    ERR_TIMEOUT = 0x0107  #263 timeout occured while waiting for interrupt
    ERR_CALLTYPE = 0x0108  #264 call type (int32 mux) is not allowed for this register
    ERR_EXCEEDSINT32 = 0x0109  #265 return value is int32 but software register exceeds the 32 bit integer range -> use 2x32 or 64
    ERR_NOWRITEALLOWED = 0x010A  #267 register cannot be written, read only
    ERR_SETUP = 0x010B  #268 the setup isn't valid
    ERR_CLOCKNOTLOCKED = 0x010C  # 268 Synchronisation to external clock failed
    ERR_CHANNEL = 0x0110  #272 Wrong number of Channel to be read out
    ERR_NOTIFYSIZE = 0x0111  #273 Notify block size isn't valid
    ERR_RUNNING = 0x0120  #288 Board is running, changes not allowed
    ERR_ADJUST = 0x0130  #304 Auto Adjust has an error
    ERR_PRETRIGGERLEN = 0x0140  #320 pretrigger length exceeds allowed values
    ERR_DIRMISMATCH = 0x0141  #321 direction of card and memory transfer mismatch
    ERR_POSTEXCDSEGMENT= 0x0142  #322 posttrigger exceeds segment size in multiple recording mode
    ERR_SEGMENTINMEM = 0x0143  #323 memsize is not a multiple of segmentsize, last segment hasn't full length
    ERR_MULTIPLEPW = 0x0144  #324 multiple pulsewidth counters used but card only supports one at the time
    ERR_NOCHANNELPWOR = 0x0145  #325 channel pulsewidth can't be OR'd
    ERR_ANDORMASKOVRLAP= 0x0146  #326 AND mask and OR mask overlap in at least one channel -> not possible
    ERR_ANDMASKEDGE = 0x0147  #327 AND mask together with edge trigger mode is not allowed
    ERR_ORMASKLEVEL = 0x0148  #328 OR mask together with level trigger mode is not allowed
    ERR_EDGEPERMOD = 0x0149  #329 All trigger edges must be simular on one module
    ERR_DOLEVELMINDIFF = 0x014A  #330 minimum difference between low output level and high output level not reached
    ERR_STARHUBENABLE = 0x014B  #331 card holding the star-hub must be active for sync
    ERR_PATPWSMALLEDGE = 0x014C  #332 Combination of pattern with pulsewidht smaller and edge is not allowed

#    -ERR_NOPCI = 0x0200  #512 No PCI bus found
#    -ERR_PCIVERSION = 0x0201  #513 Wrong PCI bus version
#    -ERR_PCINOBOARDS = 0x0202  #514 No Spectrum PCI boards found
    ERR_PCICHECKSUM = 0x0203  #515 Checksum error on PCI board
    ERR_DMALOCKED = 0x0204  #516 DMA buffer in use, try later
    ERR_MEMALLOC = 0x0205  #517 Memory Allocation error
    ERR_EEPROMLOAD = 0x0206  #518 EEProm load error, timeout occured
    ERR_CARDNOSUPPORT = 0x0207  #519 no support for that card in the library

#    -ERR_FIFOBUFOVERRUN = 0x0300  #768 Buffer overrun in FIFO mode
    ERR_FIFOHWOVERRUN = 0x0301  #769 Hardware buffer overrun in FIFO mode
    ERR_FIFOFINISHED = 0x0302  #770 FIFO transfer hs been finished. Number of buffers has been transferred
#    -ERR_FIFOSETUP = 0x0309  #777 FIFO setup not possible, transfer rate to high (max 250 MB/s)

    ERR_TIMESTAMP_SYNC = 0x0310  #784 Synchronisation to ref clock failed
    ERR_STARHUB = 0x0320  #800 Autorouting of Starhub failed

    ERR_INTERNAL_ERROR = 0xFFFF  #65535 Internal hardware error detected, please check for update





import logging
from abc import ABCMeta, abstractmethod

from qkit.core.instrument_base import Instrument


class AbstractMicrowaveSource(Instrument, metaclass=ABCMeta):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Implement parameters
        self.add_parameter('power',
                           flags=Instrument.FLAG_GETSET, units='dBm', minval=-135, maxval=30, offset=True, type=float)
        # self.add_parameter('phase',
        #    flags=Instrument.FLAG_GETSET, units='rad', minval=-numpy.pi, maxval=numpy.pi, type=types.FloatType)
        self.add_parameter('frequency',
                           flags=Instrument.FLAG_GETSET, units='Hz', minval=9e3, maxval=20e9, type=float)
        self.add_parameter('status',
                           flags=Instrument.FLAG_GETSET, type=bool)

        self.add_function('reset')
        self.add_function('get_all')

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : get all')
        self.get_power()
        #self.get_phase()
        self.get_frequency()
        self.get_status()

    @abstractmethod
    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        pass

    @abstractmethod
    def do_get_power(self):
        '''
        Reads the power of the signal from the instrument

        Input:
            None

        Output:
            ampl (float) : power in dBm
        '''
        pass

    @abstractmethod
    def do_set_power(self, amp):
        '''
        Set the power of the signal

        Input:
            amp (float) : power in dBm

        Output:
            None
        '''
        pass

    @abstractmethod
    def do_get_frequency(self):
        '''
        Reads the frequency of the signal from the instrument

        Input:
            None

        Output:
            freq (float) : Frequency in Hz
        '''
        pass

    @abstractmethod
    def do_set_frequency(self, freq):
        '''
        Set the frequency of the instrument

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        pass

    @abstractmethod
    def do_get_status(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (string) : 'On' or 'Off'
        '''
        pass

    @abstractmethod
    def do_set_status(self, status):
        '''
        Set the output status of the instrument

        Input:
            status (bool) : True: 'On' or False: 'Off'

        Output:
            None
        '''
        pass

    def off(self):
        '''
        Set status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_status(False)

    def on(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_status(True)



# Quantum machines driver, M. Spiecker 2019
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import time
import numpy as np
from qm.qua import *
from qm.QuantumMachinesManager import QuantumMachinesManager
from qkit.measure.timedomain.quantum_machine.qm_qkit_wrapper import QmQkitWrapper
from qkit.core.instrument_base import Instrument
from qkit import visa
import qkit
import logging




class virtual_VNA_quantum_machines(Instrument, QmQkitWrapper):
    """
    Class to use the quantum machines card as a VNA in the spectroscopy script.
    For the readout a interferrometric setup is assmued.
    Feel free to extend the functionality # TODO
    """

    def __init__(self, name, qm_config, mw_drive_freq_func, avg = 100):

        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])
        QmQkitWrapper.__init__(self)


        '''QM and VNA variables'''
        self.nop = 1001
        self.startfreq = 4e9
        self.stopfreq = 5e9
        self.sweeptime_averages = 1
        self.span = self.stopfreq - self.startfreq
        self.centerfreq = (self.startfreq + self.stopfreq) / 2
        self._ready = False
        self.if_end = 100e6
        self.if_start = 50e6
        self.wait_before = 2500
        self.avg = avg

        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('get_sweeptime_averages')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')
        self.mw_drive_freq_func = mw_drive_freq_func #Function that changes the Lo drive. For example: anapico.set_ch1_frequency

        #self.qmm = QuantumMachinesManager()
        self.qm_config = qm_config
        #self.qm = self.qmm.open_qm(self.qm_config)


        self.add_parameter('frequency',
                           flags=super().FLAG_GETSET, units='Hz', minval=50e6, maxval=20e9+50e6, type=float)


    '''
    def config(self):

        opx_one = "Opx"
        readout_pulse_len = 200  # attenuation 10 dB #good: 200

        QB_IF_freq = 80.1e6 + 0.08315e6
        QB_LO_freq = 5e9  # for the mixer
        RR_IF_freq = 62.5e6
        theta = 18
        wait_sig = 118
        wait_ref = 52

        config = {
            'version': 1,
            'controllers': {
                opx_one: {
                    'type': 'Opx',
                    'analog_outputs': {
                        # sweep offset of I,Q to see which combination best suppresses LO leakage
                        1: {'offset': 0.0},  # I qubit
                        2: {'offset': 0.0},  # Q qubit
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # I
                        2: {'offset': 0}  # Q
                    },
                    'digital_outputs': {
                        1: {}
                    }
                },
            },

            'elements': {

                'test': {
                    "mixInputs": {
                        "I": (opx_one, 5),
                        "Q": (opx_one, 6),
                        "mixer": "mixer_QB",
                        "lo_frequency": (QB_LO_freq)
                    },
                    'intermediate_frequency': (QB_IF_freq),
                    'operations': {
                        'test_pulse': 'test_pulse'
                    }

                },

                'test_dc': {
                    "singleInput": {
                        "port": (opx_one, 4),
                    },
                    'intermediate_frequency': (QB_IF_freq),
                    'operations': {
                        'test_dc_pulse': 'test_dc_pulse'
                    }

                },

                'qubit': {
                    "mixInputs": {
                        "I": (opx_one, 1),
                        "Q": (opx_one, 2),
                        "mixer": "mixer_QB",
                        "lo_frequency": (QB_LO_freq)
                    },
                    'intermediate_frequency': (QB_IF_freq),
                    'operations': {
                        'gaussian_pulse': 'gaussian_pulse',
                        'pi_pulse': 'pi_pulse',
                        'pi/2_pulse': 'pi/2_pulse',
                        'const': 'const_pulse',
                        'saturation_pulse': 'saturation_pulse',
                        'saturation_low_pulse': 'saturation_low_pulse'
                    }
                },

                'reference': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': RR_IF_freq,
                    'operations': {
                        'readout_ref': 'zero_pulse',
                        'readout_ref_bay': 'zero_pulse_bay'
                    },
                    "outputs": {
                        'out1': (opx_one, 1),
                    },
                    'time_of_flight': 32 + 4 * wait_ref,
                    'smearing': 0
                },

                'signal': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': RR_IF_freq,
                    'operations': {
                        'readout_sig': 'readout_pulse',
                        'readout_sig_bay': 'readout_pulse_bay'
                    },
                    "outputs": {
                        'out1': (opx_one, 2),
                    },
                    'time_of_flight': 32 + 4 * wait_sig,
                    'smearing': 0
                },

                'reference_T1': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': RR_IF_freq,
                    'operations': {
                        'readout_ref': 'zero_pulse',
                        'readout_ref_bay': 'zero_pulse_bay',
                    },
                    "outputs": {
                        'out1': (opx_one, 1),
                    },
                    'time_of_flight': 32 + 4 * wait_ref,
                    'smearing': 0
                },

                'signal_T1': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': RR_IF_freq,
                    'operations': {
                        'readout_sig': 'readout_pulse',
                        'readout_sig_bay': 'readout_pulse_bay',
                    },
                    "outputs": {
                        'out1': (opx_one, 2),
                    },
                    'time_of_flight': 32 + 4 * wait_sig,
                    'smearing': 0
                },
            },

            "pulses": {

                'test_dc_pulse': {
                    'operation': 'control',
                    'length': 100,
                    'waveforms': {
                        'single': 'const_wf',
                    },
                },

                'test_pulse': {
                    'operation': 'control',
                    'length': 100,
                    'waveforms': {
                        'I': 'const_wf',
                        'Q': 'zero_wf'
                    },
                },
                'gaussian_pulse': {
                    'operation': 'control',
                    'length': pi_pulse_len,
                    'waveforms': {
                        'I': 'gaussian_wf',
                        'Q': 'zero_wf'
                    },
                },

                'readout_pulse': {
                    'operation': 'measurement',
                    'length': readout_pulse_len,
                    'waveforms': {
                        'single': 'const_wf',
                    },
                    'digital_marker': 'ON',
                    'integration_weights': {
                        'integW_Is': 'integW_Is',
                        'integW_Qs': 'integW_Qs'
                    },
                },
                'zero_pulse': {
                    'operation': 'measurement',
                    'length': readout_pulse_len,
                    'waveforms': {
                        'single': 'zero_wf',
                    },
                    'digital_marker': 'ON',
                    'integration_weights': {
                        'integW_Ir': 'integW_Ir',
                        'integW_Qr': 'integW_Qr',
                    }
                },


                'const_pulse': {
                    'operation': 'control',
                    'length': 320,
                    'waveforms': {
                        'I': 'const_wf',
                        'Q': 'zero_wf',
                    }
                },
                'saturation_pulse': {
                    'operation': 'control',
                    'length': saturation_pulse_length,
                    'waveforms': {
                        'I': 'const_wf',
                        'Q': 'zero_wf',
                    }
                },
                'saturation_low_pulse': {
                    'operation': 'control',
                    'length': saturation_pulse_length,
                    'waveforms': {
                        'I': 'const_low_wf',
                        'Q': 'zero_wf',
                    }
                },
            },

            'waveforms': {
                'zero_wf': {
                    'type': 'constant',
                    'sample': 0
                },
                'const_wf': {
                    'type': 'constant',
                    'sample': 0.4
                },

                'gaussian_wf': {
                    'type': 'arbitrary',
                    'samples': gauss(0.45, 0.0, 10.0, pi_pulse_len)
                },
            },
            'digital_waveforms': {
                'ON': {
                    'samples': [(1, 0)]
                },
            },

            'integration_weights': {
                'integW_Ir': {
                    'cosine': [np.cos(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                    'sine': [np.sin(- theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                },
                'integW_Qr': {
                    'cosine': [np.sin(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                    'sine': [np.cos(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                },
                'integW_Is': {
                    'cosine': [np.cos(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                    'sine': [np.sin(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                },
                'integW_Qs': {
                    'cosine': [np.sin(- theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                    'sine': [np.cos(theta / 360 * np.pi)] * int(readout_pulse_len / 4),
                },
            },

            'mixers': {
                'mixer_QB': [
                    {
                        'intermediate_frequency': QB_IF_freq,
                        'lo_frequency': QB_LO_freq,
                        'correction': [1, 0, 0, 1]
                    }
                ],
            },
        }

        return config
    '''

    def get_all(self):
        pass

    def qm_program(self,f_list):

        f_list = f_list.astype(int).tolist()

        with program() as IQ_demod:

            f = declare(int)
            N = declare(int)
            I = declare(fixed)
            Q = declare(fixed)

            with for_(N, 0, N < self.avg, N + 1):

                with for_each_(f, f_list):

                    wait(self.wait_before, "resonator")
                    update_frequency('resonator', f)
                    measure("readout", "resonator", None, dual_demod.full('integ_w1_I', 'out1', 'integ_w2_I', 'out2', I)
                            , dual_demod.full('integ_w1_Q', 'out1', 'integ_w2_Q', 'out2', Q))

                    # save to client:
                    save(I, "I")
                    save(Q, "Q")
        self.program = IQ_demod


    def run_card(self):
        job = self.qm_config.qm.execute(self.program, duration_limit=0, data_limit=0)
        job.result_handles.get("I").wait_for_all_values()
        job.result_handles.get("Q").wait_for_all_values()

        I, Q =  job.result_handles.get("I").fetch_all()["value"], job.result_handles.get("Q").fetch_all()["value"]

        I = np.reshape(I, (self.avg, len(self.qm_freqs)))
        Q = np.reshape(Q, (self.avg, len(self.qm_freqs)))
        I_avg = np.mean(I, axis=0)
        Q_avg = np.mean(Q, axis=0)

        while(len(I_avg) < self.nop_perblock):
            I_avg = np.append(I_avg,[0])
        while (len(Q_avg) < self.nop_perblock):
            Q_avg = np.append(Q_avg, [0])

        return I_avg, Q_avg

    def set_startfreq(self, startfreq):
        self.startfreq = startfreq
        self.span = self.stopfreq - self.startfreq
        self.centerfreq = (self.stopfreq + self.startfreq) / 2

    def get_startfreq(self):
        return self.startfreq

    def set_stopfreq(self, stopfreq):
        self.stopfreq = stopfreq
        self.span = self.stopfreq - self.startfreq
        self.centerfreq = (self.stopfreq + self.startfreq) / 2

    def get_stopfreq(self):
        return self.stopfreq

    def set_centerfreq(self, centerfreq):
        self.centerfreq = centerfreq
        self.startfreq = centerfreq - self.span / 2
        self.stopfreq = centerfreq + self.span / 2

    def get_centerfreq(self):
        return self.centerfreq

    def get_span(self):
        return self.span

    def set_span(self, span):
        self.span = span
        self.startfreq = self.centerfreq - span / 2
        self.stopfreq = self.centerfreq + span / 2

    def get_nop(self):
        return self.nop

    def set_nop(self, nop):
        self.nop = nop

    def get_freqpoints(self):
        return np.linspace(self.startfreq, self.stopfreq, self.nop)

    def get_sweeptime_averages(self,query=True):
        return self.sweeptime_averages

    def get_sweeptime(self,query=True):
        return 1

    def get_averages(self):
        return 1

    def get_Average(self):
        return False

    def get_tracedata(self, RealImag=None):

        '''Logic to define Sweep steps'''
        self.frequencies, delta_f = np.linspace(self.startfreq, self.stopfreq, self.nop, endpoint=True, retstep=True)
        nop_perblock = int((self.if_end - self.if_start) / delta_f)
        self.qm_freqs = np.arange(nop_perblock) * delta_f + self.if_start
        mw_freqlist = np.arange(self.startfreq, self.stopfreq, nop_perblock * delta_f) - self.if_start

        if(self.nop > nop_perblock*len(mw_freqlist)):
            mw_freqlist = np.append(mw_freqlist, mw_freqlist[-1] + nop_perblock * delta_f)

        #DEBUG__
        self.nop_perblock = nop_perblock
        self.mw_freqlist = mw_freqlist


        '''Actual tracedata getting'''
        I = np.empty(nop_perblock*len(mw_freqlist))
        Q = np.empty(nop_perblock*len(mw_freqlist))
        self.qm_program(self.qm_freqs)

        for i, mw in enumerate(mw_freqlist):
            self.mw_drive_freq_func(mw)
            time.sleep(0.5)
            I[i*nop_perblock:(i+1)*nop_perblock], Q[i*nop_perblock:(i+1)*nop_perblock] = self.run_card()

        return I[:self.nop-nop_perblock*len(mw_freqlist)],Q[:self.nop-nop_perblock*len(mw_freqlist)]


        '''
        def S21_notch(f, fr, Ql, Qc):
            return 1. - Ql / Qc / (1. + 2j * Ql * (f - fr) / fr)

        Qi = np.random.normal(2000, 500)
        Qc = np.random.normal(1500, 500)  # for a nice signal the resonance is overcoupled
        fr = np.random.normal((self.stopfreq + self.startfreq) / 2, self.span / 8)
        Ql = Qi * Qc / (Qi + Qc)
        S21_data = S21_notch(self.get_freqpoints(), fr, Ql, Qi) + np.random.normal(0., 0.01, self.nop)

        if not RealImag:
            amp = np.abs(S21_data)
            pha = np.angle(S21_data)
            return amp, pha
        else:
            I = S21_data.real
            Q = S21_data.imag
            return I, Q'''

    def get_all(self):
        pass

    def pre_measurement(self):
        pass

    def post_measurement(self):
        pass

    def start_measurement(self):
        pass

    def ready(self):
        self._ready = not self._ready
        return self._ready

    def avg_clear(self):
        pass

    def gauss(amplitude, mu, sigma, length):
        t = np.linspace(-length / 2, length / 2, length)
        gauss_wave = amplitude * np.exp(-((t - mu) ** 2) / (2 * sigma ** 2))
        return [float(x) for x in gauss_wave]

    def IQ_imbalance_corr(g, phi):
        c = np.cos(phi)
        s = np.sin(phi)
        N = 1 / ((1 - g ** 2) * (2 * c ** 2 - 1))
        return [float(N * x) for x in [(1 - g) * c, (1 + g) * s,
                                       (1 - g) * s, (1 + g) * c]]